'use strict';

// FBP protocol dependent code
var common, debug, getComponents;

common = require('./common');

debug = require('debug')('fbp-spec:protocol');

exports.sendGraph = function (runtime, graph, callback) {
  var connection, graphId, i, iip, len, main, name, pendingPorts, priv, process, pub, ref, ref1, ref2, ref3, _waitForPorts;
  main = false; // this is a component?
  if (!graph) {
    return callback(new Error("Graph not defined"));
  }
  graphId = graph.name || graph.properties.id;
  if (!graphId) {
    graphId = 'fixture.' + common.randomString(10);
  }
  pendingPorts = {
    in: [],
    out: []
  };
  runtime.sendGraph('clear', {
    id: graphId,
    name: graph.name,
    main: main,
    library: graph.properties.project || '',
    icon: graph.properties.icon || '',
    description: graph.properties.description || ''
  });
  ref = graph.processes;
  for (name in ref) {
    process = ref[name];
    debug('adding node', name, process.component);
    runtime.sendGraph('addnode', {
      id: name,
      component: process.component,
      metadata: process.metadata,
      graph: graphId
    });
  }
  ref1 = graph.connections;
  for (i = 0, len = ref1.length; i < len; i++) {
    connection = ref1[i];
    if (connection.src != null) {
      debug('connecting edge', connection);
      runtime.sendGraph('addedge', {
        src: {
          node: connection.src.process,
          port: connection.src.port
        },
        tgt: {
          node: connection.tgt.process,
          port: connection.tgt.port
        },
        metadata: connection.metadata != null,
        graph: graphId
      });
    } else {
      iip = connection;
      debug('adding IIP', iip);
      runtime.sendGraph('addinitial', {
        src: {
          data: iip.data
        },
        tgt: {
          node: iip.tgt.process,
          port: iip.tgt.port
        },
        metadata: iip.metadata,
        graph: graphId
      });
    }
  }
  if (graph.inports) {
    ref2 = graph.inports;
    for (pub in ref2) {
      priv = ref2[pub];
      debug('exporting inport', pub);
      runtime.sendGraph('addinport', {
        public: pub,
        node: priv.process,
        port: priv.port,
        graph: graphId
      });
      pendingPorts.in.push(pub);
    }
  }
  if (graph.outports) {
    ref3 = graph.outports;
    for (pub in ref3) {
      priv = ref3[pub];
      debug('exporting outport', pub);
      runtime.sendGraph('addoutport', {
        public: pub,
        node: priv.process,
        port: priv.port,
        graph: graphId
      });
      pendingPorts.out.push(pub);
    }
  }
  _waitForPorts = function waitForPorts(_ref) {
    var command = _ref.command,
        payload = _ref.payload;

    var collection;
    if (command !== 'addinport' && command !== 'addoutport') {
      return;
    }
    debug('received port', payload.public);
    if (command === 'addinport') {
      collection = pendingPorts.in;
    } else {
      collection = pendingPorts.out;
    }
    if (collection.indexOf(payload.public) === -1) {
      debug('received unknown port', payload.public);
      return;
    }
    collection.splice(collection.indexOf(payload.public), 1);
    if (pendingPorts.in.length || pendingPorts.out.length) {
      return;
    }
    runtime.removeListener('graph', _waitForPorts);
    return callback(null, graphId);
  };
  debug('sendGraph waiting for updated exported ports');
  return runtime.on('graph', _waitForPorts);
};

exports.startNetwork = function (runtime, graphId, callback) {
  var _waitForStarted;
  debug('startnetwork', graphId);
  _waitForStarted = function waitForStarted(status) {
    debug('start: runtime status change', status);
    if (status.started) {
      runtime.removeListener('execution', _waitForStarted);
      return callback(null);
    }
  };
  runtime.on('execution', _waitForStarted);
  return runtime.sendNetwork('start', {
    graph: graphId
  });
};

exports.stopNetwork = function (runtime, graphId, callback) {
  var _waitForStopped;
  debug('stopnetwork', graphId);
  _waitForStopped = function waitForStopped(status) {
    debug('stop: runtime status change', status);
    if (!status.running) {
      runtime.removeListener('execution', _waitForStopped);
      return callback(null);
    }
  };
  runtime.on('execution', _waitForStopped);
  return runtime.sendNetwork('stop', {
    graph: graphId
  });
};

exports.sendPackets = function (client, graphId, packets, callback) {
  var payload, port;
  debug('sendpackets', graphId, packets);
  for (port in packets) {
    payload = packets[port];
    client.sendRuntime('packet', {
      event: 'data',
      port: port,
      payload: payload,
      graph: graphId
    });
  }
  return callback(null);
};

exports.getComponents = getComponents = function getComponents(client, callback) {
  var components, _gotComponent;
  debug('get components');
  components = {};
  _gotComponent = function gotComponent(msg) {
    var command, payload;
    command = msg.command;
    payload = msg.payload;

    debug('got component?', command);
    if (command === 'component') {
      return components[payload.name] = payload;
    } else if (command === 'componentsready') {
      client.removeListener('component', _gotComponent);
      return callback(null, components);
    }
  };
  client.on('component', _gotComponent);
  return client.sendComponent('list', {});
};

exports.getCapabilities = function (client, callback) {
  var def, _onCapabilities, ref;
  def = client.definition;
  if (def != null ? (ref = def.capabilities) != null ? ref.length : void 0 : void 0) {
    return callback(null, def.capabilities);
  }
  _onCapabilities = function onCapabilities(capabilities) {
    client.removeListener('capabilities', _onCapabilities);
    return callback(null, capabilities, def);
  };
  return client.on('capabilities', _onCapabilities);
};

exports.getComponentTests = function (client, callback) {
  var complete, expectResponses, gotComponent, responses, tests;
  debug('get component tests');
  responses = 0;
  expectResponses = 0;
  tests = {};
  gotComponent = function gotComponent(msg) {
    var command, payload;
    command = msg.command;
    payload = msg.payload;

    responses += 1;
    debug('got component source?', command, payload.name, payload.tests != null, responses, expectResponses);
    if (command !== 'source') {
      return;
    }
    if (payload.tests != null) {
      tests[payload.name] = payload.tests; // not all components have tests
    }
    if (responses === expectResponses) {
      debug('got all component sources', Object.keys(tests).length);
      return complete(null, tests);
    }
  };
  complete = function complete(err, tests) {
    client.removeListener('component', gotComponent);
    return callback(err, tests);
  };
  return getComponents(client, function (err, components) {
    var componentNames, i, len, name, results;
    if (err) {
      return complete(err);
    }
    componentNames = Object.keys(components);
    expectResponses = componentNames.length;
    if (expectResponses === 0) {
      return complete(null, tests);
    }
    debug('retrieving sources for', expectResponses);
    client.on('component', gotComponent);
    results = [];
    for (i = 0, len = componentNames.length; i < len; i++) {
      name = componentNames[i];
      results.push(client.sendComponent('getsource', {
        name: name
      }));
    }
    return results;
  });
};